/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software License
 * version 1.1, a copy of which has been included with this distribution in
 * the LICENSE file.
 */
package org.apache.avalon.excalibur.io;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * This class provides basic facilities for manipulating io streams.
 *
 * @author <a href="mailto:donaldp@apache.org">Peter Donald</a>
 */
public final class IOUtil
{
    /**
     * Private constructor to prevent instantiation.
     */
    private IOUtil()
    {
    }

    public static void shutdownStream( final OutputStream output )
    {
        if( null == output )
        {
            return;
        }

        try { output.close(); }
        catch( final IOException ioe ) {}
    }

    public static void shutdownStream( final InputStream input )
    {
        if( null == input )
        {
            return;
        }

        try { input.close(); }
        catch( final IOException ioe ) {}
    }

    /**
     * Copy stream-data from source to destination. This method does not
     * buffer, flush or close the streams, as to do so would require making
     * non-portable assumptions about the streams' origin and further use. If
     * you wish to perform a buffered copy, use {@link #bufferedCopy}.
     */
    public static void copy( final InputStream input, final OutputStream output )
        throws IOException
    {
        final int BUFFER_SIZE = 1024 * 4;
        final byte[] buffer = new byte[ BUFFER_SIZE ];

        while( true )
        {
            final int count = input.read( buffer, 0, BUFFER_SIZE );
            if( -1 == count )
            {
                break;
            }
            
            // write out those same bytes
            output.write( buffer, 0, count );
        }

        //needed to flush cache
        //output.flush();
    }

    /**
     * Copy stream-data from source to destination, with buffering.
     * This is equivalent to passing {@link #copy} a {@link
     * java.io.BufferedInputStream} and {@link java.io.BufferedOuputStream} to
     * {@link #copy}, and flushing the output stream afterwards. The streams
     * are not closed after the copy.
     */
    public static void bufferedCopy( final InputStream source, final OutputStream destination )
        throws IOException
    {
        final BufferedInputStream input = new BufferedInputStream( source );
        final BufferedOutputStream output = new BufferedOutputStream( destination );
        copy( input, output );
        output.flush();
    }
}
