/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software License
 * version 1.1, a copy of which has been included with this distribution in
 * the LICENSE file.
 */
package org.apache.avalon.excalibur.i18n;

/** JDK classes **/
import java.util.Hashtable;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.Vector;

import org.xml.sax.SAXParseException;

import org.apache.log.LogKit;
import org.apache.log.Logger;

/**
 * This is the XMLResourceBundleFactory, the method for getting and
 * creating XMLResourceBundles.
 *
 * FIXME: We need to make this an instantiated class that is configurable.
 *
 * @author <a href="mailto:mengelhart@earthtrip.com">Mike Engelhart</a>
 * @author <a href="mailto:neeme@one.lv">Neeme Praks</a>
 * @author <a href="mailto:oleg@one.lv">Oleg Podolsky</a>
 * @version $Id: XMLResourceBundleFactory.java,v 1.2 2001/05/01 13:41:42 bloritsch Exp $
 */

public class XMLResourceBundleFactory
{
    /**
     * Boolean stating whether to cache resourcebundles on startup
     * or not.  You can override this by extending the class, or have
     * a static setter in the same package.
     */
    protected static boolean cacheAtStartup = true;

    protected static Hashtable cache = new Hashtable();
    protected static Hashtable cache_not_found = new Hashtable();
    protected static String directory;
    protected static Logger logger = LogKit.getLoggerFor("XMLResourceBundle");

    public static XMLResourceBundle getBundle(String name) throws MissingResourceException
    {
        return getBundle(name, Locale.getDefault());
    }

    public static XMLResourceBundle getBundle(String name, Locale loc) throws MissingResourceException
    {
        return getBundle(name, loc, cacheAtStartup);
    }

    public static XMLResourceBundle getBundle(String name, Locale loc, boolean cacheAtStartup) throws MissingResourceException
    {
        XMLResourceBundle bundle = _getBundle(name, loc, cacheAtStartup);
        if (bundle == null)
            throw new MissingResourceException("Unable to locate resource: " + name, XMLResourceBundleFactory.class.getName(), "");
        else
            bundle.setLogger(logger);
            return bundle;
    }

    private static XMLResourceBundle _getBundle(String name, Locale loc)
    {
        return _getBundle(name, loc, cacheAtStartup);
    }

    protected static XMLResourceBundle getParentBundle(String name, Locale loc)
    {
        return getParentBundle(name, loc, cacheAtStartup);
    }

    protected static XMLResourceBundle getParentBundle(String name, Locale loc, boolean cacheAtStartup)
    {
        return _getBundle(name, getParentLocale(loc), cacheAtStartup);
    }

    public static XMLResourceBundle getBundle(String fileName, String localeName) throws MissingResourceException
    {
        return getBundle(fileName, new Locale(localeName, localeName) );
    }

    public static XMLResourceBundle getBundleFromFilename(String bundleName) throws MissingResourceException
    {
        return getBundleFromFilename(bundleName, cacheAtStartup);
    }

    public static XMLResourceBundle getBundleFromFilename(String name, boolean cacheAtStartup) throws MissingResourceException
    {
        XMLResourceBundle bundle = _getBundle(name, null, cacheAtStartup);
        if (bundle == null)
            throw new MissingResourceException("Unable to locate resource: " + name, "XMLResourceBundleFactory", "");
        else
            bundle.setLogger(logger);
            return bundle;
    }

    private static XMLResourceBundle _getBundle(String name, Locale loc, boolean cacheAtStartup)
    {
        if (logger.isDebugEnabled()) logger.debug("_getBundle: " + name + ", locale " + loc);
        String bundleName = getBundleName(name, loc);
        XMLResourceBundle bundle = getCachedBundle(bundleName);
        String noBundle = getNotFoundBundle(bundleName);
        if (bundle == null && noBundle == null)
        {
            if (logger.isDebugEnabled()) logger.debug("Not found in cache, loading: " + bundleName);
            synchronized(XMLResourceBundleFactory.class)
            {
                bundle = getCachedBundle(bundleName);
                noBundle = getNotFoundBundle(bundleName);
                if (bundle == null && noBundle == null)
                {
                    if (logger.isDebugEnabled()) logger.debug("Still not found in cache, loading: " + bundleName);
                    bundle = _loadBundle(name, bundleName, loc, cacheAtStartup);
                    Locale parentLoc = loc;
                    String parentBundleName;
                    while (bundle == null && parentLoc != null && !parentLoc.getLanguage().equals(""))
                    {
                        if (logger.isDebugEnabled()) logger.debug("Still haven't found bundle, trying: " + bundleName);
                        parentLoc = getParentLocale(parentLoc);
                        parentBundleName = getBundleName(name, parentLoc);
                        bundle = _loadBundle(name, parentBundleName, parentLoc, cacheAtStartup);
                        updateCache(parentBundleName, bundle);
                    }
                    updateCache(bundleName, bundle);
                }
            }
        }
        return bundle;
    }

    private static XMLResourceBundle _loadBundle(String name, String bundleName, Locale loc, boolean cacheAtStartup)
    {
        if (logger.isDebugEnabled()) logger.debug("Trying to load bundle: " + name + ", locale " + loc + ", filename " + bundleName);
        XMLResourceBundle bundle = null;
        XMLResourceBundle parentBundle = null;
        try
        {
            if (loc != null && !loc.getLanguage().equals(""))
                parentBundle = getParentBundle(name, loc);
            bundle = new XMLResourceBundle(name, bundleName, loc, parentBundle, cacheAtStartup);
            if (bundle != null) bundle.setLogger(logger);
        }
        catch (SAXParseException e)
        {
            if (logger.isInfoEnabled()) logger.info("Resource loading failed: " + e.getMessage());
        }
        catch (Exception e)
        {
            logger.error("Error while loading resource: " + name + ", locale " + loc + ", bundleName " + bundleName, e);
        }
        return bundle;
    }

    // this method returns the next locale up the parent hierarchy
    // e.g.; the parent of new Locale("en","us","mac")
    // would be new Locale("en", "us", "");
    protected static Locale getParentLocale(Locale loc)
    {
        Locale newloc;
        if (loc.getVariant().equals(""))
        {
            if (loc.getCountry().equals(""))
                newloc = new Locale("","","");
            else
                newloc = new Locale(loc.getLanguage(), "", "");
        }
        else
            newloc = new Locale(loc.getLanguage(), loc.getCountry(), "");

        return newloc;
    }

    protected static String getBundleName(String name, Locale loc)
    {
        StringBuffer sb = new StringBuffer( getDirectory() );
        if (loc != null)
        {
            String lang = loc.getLanguage();
            String country = loc.getCountry();
            String variant = loc.getVariant();

            if (lang.length() > 0) sb.append("/").append(lang);
            if (country.length() > 0) sb.append("/").append(country);
            if (variant.length() > 0) sb.append("/").append(variant);
        }
        sb.append("/").append(name).append(".xml");

        String result = sb.toString();
        if (logger.isDebugEnabled()) logger.debug("Resolving bundle name to file name: " + name + ", locale " + loc + " --> " + result);
        return result;
    }

    protected static XMLResourceBundle getCachedBundle(String bundleName)
    {
        XMLResourceBundle bundle = (XMLResourceBundle)(cache.get(bundleName));
        if (bundle != null)
        {
            if (logger.isDebugEnabled()) logger.debug("Returning from cache: " + bundleName);
        }
        else
        {
            if (logger.isDebugEnabled()) logger.debug("Not found in cache: " + bundleName);
        }
        return bundle;
    }

    protected static String getNotFoundBundle(String bundleName)
    {
        String result = (String)(cache_not_found.get(bundleName));
        if (result != null)
        {
            if (logger.isDebugEnabled()) logger.debug("Returning from not_found_cache: " + bundleName);
        }
        else
        {
            if (logger.isDebugEnabled()) logger.debug("Not found in not_found_cache: " + bundleName);
        }
        return result;
    }

    protected static void updateCache(String bundleName, XMLResourceBundle bundle)
    {
        if (bundle == null)
            updateNotFoundCache(bundleName);
        else
            updateFoundCache(bundleName, bundle);
    }

    protected static void updateFoundCache(String bundleName, XMLResourceBundle bundle)
    {
        if (logger.isDebugEnabled()) logger.debug("Updating cache: " + bundleName);
        cache.put(bundleName, bundle);
    }

    protected static void updateNotFoundCache(String bundleName)
    {
        if (logger.isDebugEnabled()) logger.debug("Updating not_found_cache: " + bundleName);
        cache_not_found.put(bundleName, bundleName);
    }

    public static void setDirectory(String dir)
    {
        directory = dir;
    }

    public static String getDirectory()
    {
        return (directory != null ? directory : "");
    }
}
