/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software License
 * version 1.1, a copy of which has been included with this distribution in
 * the LICENSE file.
 */
package org.apache.avalon.framework.util.test;

import java.util.List;
import org.apache.avalon.framework.util.CircularDependencyException;
import org.apache.avalon.framework.util.DependencyGraph;
import org.apache.testlet.AbstractTestlet;

/**
 *
 * @author <a href="mailto:donaldp@apache.org">Peter Donald</a>
 */
public final class DependencyGraphTestlet
    extends AbstractTestlet
{
    protected final String[][]        DEPENDENCY_TREE =
    {
        { "A" }, { "B", "C", "D" },
        { "B" }, { "E" },
        { "C" }, {},
        { "D" }, { "F" },
        { "E" }, { "F" },
        { "F" }, { },
        { "G" }, { "H" },
        { "H" }, { "G" },
        { "I" }, { "I" }
    };

    protected DependencyGraph       m_graph;

    public void initialize()
    {
        m_graph = new DependencyGraph();

        for( int i = 0; i < DEPENDENCY_TREE.length; i += 2 )
        {
            m_graph.add( DEPENDENCY_TREE[ i ][0], DEPENDENCY_TREE[ i + 1 ] );
        }
    }

    protected boolean contains( final DependencyGraph.Dependency[] list, final String name )
    {
        for( int i = 0; i < list.length; i++ )
        {
            if( list[ i ].getName().equals( name ) )
            {
                return true;
            }
        }

        return false;
    }

    public void testNoDependency()
        throws Exception
    {
        final DependencyGraph.Dependency[] list = m_graph.getDependencyList( "F" );

        assertEquality( "Graph for F", list.length, 1 );
        assertEquality( "Graph for F[1]", list[ 0 ].getName(), "F" );
        assertEquality( "Graph for F[1]", list[ 0 ].getRequiredBy(), null );
    }

    public void test1LevelDependency()
        throws Exception
    {
        final DependencyGraph.Dependency[] list = m_graph.getDependencyList( "E" );

        assertEquality( "Graph for E", list.length, 2 );
        assert( "Graph for E[1]", contains( list, "E" ) );
        assert( "Graph for E[2]", contains( list, "F" ) );
    }

    public void test2LevelDependency()
        throws Exception
    {
        final DependencyGraph.Dependency[] list = m_graph.getDependencyList( "B" );

        assertEquality( "Graph for E", list.length, 3 );
        assert( "Graph for E[1]", contains( list, "E" ) );
        assert( "Graph for E[2]", contains( list, "F" ) );
        assert( "Graph for E[3]", contains( list, "B" ) );
    }

    public void testNLevelDependency()
        throws Exception
    {
        final DependencyGraph.Dependency[] list = m_graph.getDependencyList( "A" );

        assertEquality( "Graph for A", list.length, 6 );
        assert( "Graph for A[1]", contains( list, "A" ) );
        assert( "Graph for A[2]", contains( list, "B" ) );
        assert( "Graph for A[3]", contains( list, "C" ) );
        assert( "Graph for A[4]", contains( list, "D" ) );
        assert( "Graph for A[5]", contains( list, "E" ) );
        assert( "Graph for A[6]", contains( list, "F" ) );
    }

    public void testAllowableCircularDependency()
        throws Exception
    {
        m_graph.setAllowCircularity( true );
        final DependencyGraph.Dependency[] list = m_graph.getDependencyList( "G" );

        assertEquality( "Graph for G", list.length, 2 );
        assert( "Graph for G[1]", contains( list, "G" ) );
        assert( "Graph for H[2]", contains( list, "H" ) );
    }

    public void testUnallowableCircularDependency()
        throws Exception
    {
        try
        {
            m_graph.setAllowCircularity( false );
            m_graph.getDependencyList( "G" );
        }
        catch( final CircularDependencyException cde )
        {
            return;
        }

        fail( "Expected CircularDependencyException" );
    }
}
